#!/bin/bash
set -e

# Build example
# ./configure --prefix=/tmp/pg --enable-debug --enable-cassert >/dev/null && make -j8 >/dev/null && make -j8 -C contrib >/dev/null && make install && make install -C contrib

# Choose server id
if [ -z $1 ]; then
  SID=01
else
  SID=$1
fi

# Set global variables
NUM_INSTALLCHECKS=100
NUM_ITERATIONS=5

SD="$( cd "$( dirname "${BASH_SOURCE[0]}" )" && pwd )"

PGPREFIX=/tmp/pg/
SRCCOPYDIR=/tmp/t/

PGDB=`pwd`/tmpdb
PGBIN=$PGPREFIX/bin
export PATH="$PGBIN:$PATH"
export LD_LIBRARY_PATH="$PGPREFIX/lib"

export PGHOST=127.0.0.1,127.0.0.1,127.0.0.1
export PGPORT=5432
export PGUSER=postgres
export PGDATABASE=test
export PGCONNECT_TIMEOUT=120

# Stop previous instance in case it is left from a previous run
$PGBIN/pg_ctl -w -t 5 -D "$PGDB" stop || true

# And clear data and logs
rm -rf "$PGDB" server*.log rm -rf installcheck$SID-*.log || true

# Prepare template directory
rm -rf $SRCCOPYDIR/template || true
mkdir $SRCCOPYDIR/template
cp * $SRCCOPYDIR/template >/dev/null 2>&1 || true
echo "exit 0" > $SRCCOPYDIR/template/config.status
mkdir $SRCCOPYDIR/template/src
cp src/* $SRCCOPYDIR/template/src/ >/dev/null 2>&1 || true
cp -r src/test $SRCCOPYDIR/template/src/
cp -r src/makefiles $SRCCOPYDIR/template/src
sed -e 's/^all:/#all:/' -i $SRCCOPYDIR/template/src/test/regress/GNUmakefile

for m in `find . -name Makefile`; do
  mkdir -p $SRCCOPYDIR/template/`dirname "$m"`
  cp $m $SRCCOPYDIR/template/$m
done

for c in `find contrib -name '*.conf'`; do
    cp $c $SRCCOPYDIR/template/$c
done

cp contrib/contrib-global.mk $SRCCOPYDIR/template/contrib/

# Avoid rebuilding
sed -e "s/all: all-lib/all:/" -i $SRCCOPYDIR/template/src/interfaces/libpq/Makefile
sed -e "s/all:.*\$/all:/" -i $SRCCOPYDIR/template/src/port/Makefile
sed -e "s/all:.*\$/all:/" -i $SRCCOPYDIR/template/src/common/Makefile
sed -e "s/all:.*\$/all:/" -i $SRCCOPYDIR/template/src/test/isolation/Makefile

# Copy whole ecpg test
if [ -f src/interfaces/ecpg/test/pg_regressXX ]; then
  cp -r src/interfaces/ecpg/test $SRCCOPYDIR/template/src/interfaces/ecpg/
else
  sed -e "s/^installcheck:/noinstallcheck:/" -i $SRCCOPYDIR/template/src/interfaces/ecpg/test/Makefile
fi

for d in `find . -type d -name sql -o -name input`; do
  mkdir -p $SRCCOPYDIR/template/`dirname "$d"`
  cp -r $d $SRCCOPYDIR/template/$d
  [ -d $d/../data ] && cp -r $d/../data $SRCCOPYDIR/template/$d/../data
  [ -d $d/../expected ] && cp -r $d/../expected $SRCCOPYDIR/template/$d/../expected
  [ -d $d/../output ] && cp -r $d/../output $SRCCOPYDIR/template/$d/../output
done

# Prepare installcheck directories
rm -rf  $SRCCOPYDIR/ic$SID-* || true
for c in `seq $NUM_INSTALLCHECKS`; do
  mkdir $SRCCOPYDIR/ic$SID-$c/
  cp -r $SRCCOPYDIR/template/* $SRCCOPYDIR/ic$SID-$c/

  # Fixing tests

  cp3=`printf "%-3s" $c`
  c03=`printf "%03d" $c`

  sed -e "s/pl_regression/pl_regress$c03/; s/contrib_regression/contrib_regress$c03/; s/isolation_regression/isolation_regress$c03/" -i $SRCCOPYDIR/ic$SID-$c/src/Makefile.global

  sed -e "s/regression/regress$cp3/" -i $SRCCOPYDIR/ic$SID-$c/src/test/regress/expected/updatable_views.out
  sed -e "s/regression/regress$cp3/" -i $SRCCOPYDIR/ic$SID-$c/src/test/regress/expected/sequence.out

  sed -e "s/ regression/ regress$c/" -i $SRCCOPYDIR/ic$SID-$c/src/test/regress/sql/publication.sql \
    $SRCCOPYDIR/ic$SID-$c/src/test/regress/expected/publication.out

  sed -e "s/ regression/ regress$c/" -i $SRCCOPYDIR/ic$SID-$c/src/test/regress/sql/dependency.sql \
    $SRCCOPYDIR/ic$SID-$c/src/test/regress/expected/dependency.out

  [ -f $SRCCOPYDIR/ic$SID-$c/src/test/regress/sql/atx.sql ] && \
  sed -e "s/\\\\c regression\$/\\\\c regress$c/" -i $SRCCOPYDIR/ic$SID-$c/src/test/regress/sql/atx.sql \
    $SRCCOPYDIR/ic$SID-$c/src/test/regress/expected/atx.out \
    $SRCCOPYDIR/ic$SID-$c/src/test/regress/expected/atx_*.out

  sed -e "s/SELECT obj_description(s.oid, 'pg_subscription') FROM pg_subscription s;/SELECT obj_description FROM (SELECT obj_description(s.oid, 'pg_subscription') FROM pg_subscription s) d WHERE obj_description IS NOT NULL LIMIT 1;/" -i \
    $SRCCOPYDIR/ic$SID-$c/src/test/regress/sql/subscription.sql \
    $SRCCOPYDIR/ic$SID-$c/src/test/regress/expected/subscription.out

  sed -e "s/SELECT pg_notification_queue_usage();/SELECT round(pg_notification_queue_usage() * 1000) \\/ 1000 AS pg_notification_queue_usage/" -i \
    $SRCCOPYDIR/ic$SID-$c/src/test/regress/sql/async.sql \
    $SRCCOPYDIR/ic$SID-$c/src/test/regress/expected/async.out

  sed -e "s/deptype NOT IN ('a', 'o', 'p', 'r')/deptype NOT IN ('a', 'o', 'p', 'r', 't')/" -i \
    $SRCCOPYDIR/ic$SID-$c/src/test/regress/sql/misc_sanity.sql \
    $SRCCOPYDIR/ic$SID-$c/src/test/regress/expected/misc_sanity.out

  sed -e "s/test_file/t${c03}_file/g" -i \
    $SRCCOPYDIR/ic$SID-$c/contrib/adminpack/sql/adminpack.sql \
    $SRCCOPYDIR/ic$SID-$c/contrib/adminpack/expected/adminpack.out

  # Add alternate expected output for some unstable tests
  patch $SRCCOPYDIR/ic$SID-$c/src/test/regress/expected/create_index.out -p1 $SD/create_index.patch -o $SRCCOPYDIR/ic$SID-$c/src/test/regress/expected/create_index_1.out >/dev/null

  patch $SRCCOPYDIR/ic$SID-$c/src/test/regress/expected/subselect.out -p1 $SD/subselect.patch -o $SRCCOPYDIR/ic$SID-$c/src/test/regress/expected/subselect_1.out >/dev/null

  patch $SRCCOPYDIR/ic$SID-$c/src/test/regress/expected/incremental_sort.out -p1 $SD/incremental_sort.patch -o $SRCCOPYDIR/ic$SID-$c/src/test/regress/expected/incremental_sort_1.out >/dev/null

  patch $SRCCOPYDIR/ic$SID-$c/src/test/regress/expected/union.out -p1 $SD/union.patch -o $SRCCOPYDIR/ic$SID-$c/src/test/regress/expected/union_1.out >/dev/null

  patch $SRCCOPYDIR/ic$SID-$c/src/test/regress/expected/join.out -p1 $SD/join.patch -o $SRCCOPYDIR/ic$SID-$c/src/test/regress/expected/join_1.out >/dev/null

  patch $SRCCOPYDIR/ic$SID-$c/src/test/regress/expected/aggregates.out -p1 $SD/aggregates.patch -o $SRCCOPYDIR/ic$SID-$c/src/test/regress/expected/aggregates_1.out >/dev/null

  patch $SRCCOPYDIR/ic$SID-$c/src/test/regress/expected/rowtypes.out -p1 $SD/rowtypes.patch -o $SRCCOPYDIR/ic$SID-$c/src/test/regress/expected/rowtypes_1.out >/dev/null

  patch $SRCCOPYDIR/ic$SID-$c/src/test/isolation/expected/deadlock-parallel.out -p1 $SD/deadlock-parallel.patch -o $SRCCOPYDIR/ic$SID-$c/src/test/isolation/expected/deadlock-parallel_1.out >/dev/null

  patch $SRCCOPYDIR/ic$SID-$c/src/test/isolation/expected/async-notify.out -p1 $SD/async-notify.patch -o $SRCCOPYDIR/ic$SID-$c/src/test/isolation/expected/async-notify_1.out >/dev/null

  patch $SRCCOPYDIR/ic$SID-$c/contrib/pageinspect/expected/page.out -p1 $SD/pageinspect.patch -o $SRCCOPYDIR/ic$SID-$c/contrib/pageinspect/expected/page_1.out >/dev/null

  # Disable test foreign_data as not fixable by a simple sed
  sed -e "s/foreign_data //" -i $SRCCOPYDIR/ic$SID-$c/src/test/regress/parallel_schedule
  sed -e "s/test: foreign_data\$//" -i $SRCCOPYDIR/ic$SID-$c/src/test/regress/serial_schedule

  # Disable test password as not fixable by a simple sed (md5 password contains the username (regress_passwd))
  sed -e "s/password //" -i $SRCCOPYDIR/ic$SID-$c/src/test/regress/parallel_schedule
  sed -e "s/test: password\$//" -i $SRCCOPYDIR/ic$SID-$c/src/test/regress/serial_schedule

  [ -f $SRCCOPYDIR/ic$SID-$c/contrib/passwordcheck/Makefile ] && sed -e 's/^REGRESS\s= /NOREGRESS = /' -i $SRCCOPYDIR/ic$SID-$c/contrib/passwordcheck/Makefile

  # Disable tests select_parallel, stats and advisory_locks as unstable in parallel runs
  sed -e "s/select_parallel //" -i $SRCCOPYDIR/ic$SID-$c/src/test/regress/parallel_schedule
  sed -e "s/test: select_parallel\$//" -i $SRCCOPYDIR/ic$SID-$c/src/test/regress/serial_schedule
  sed -e "s/advisory_lock //" -i $SRCCOPYDIR/ic$SID-$c/src/test/regress/parallel_schedule
  sed -e "s/test: advisory_lock\$//" -i $SRCCOPYDIR/ic$SID-$c/src/test/regress/serial_schedule
  sed -e "s/test: stats\$//" -i $SRCCOPYDIR/ic$SID-$c/src/test/regress/parallel_schedule
  sed -e "s/test: stats\$//" -i $SRCCOPYDIR/ic$SID-$c/src/test/regress/serial_schedule

  # Fix regress entities
  find  $SRCCOPYDIR/ic$SID-$c/ \( -name *.sql -o -name *.source -o -name *.out -o -name *.spec \) -exec sed -e \
"s/regress_\(tblspace\)/regr${c03}_\1/g; "\
"s/regress_\(vacuum\)/regr${c03}_\1/g; "\
"s/regress_\(other_partitioned_fk_owner\)/regr${c03}_\1/g; "\
"s/regress_\(testsub\)/regr${c03}_\1/g; "\
"s/regress_\(regrole_test\)/regr${c03}_\1/g; "\
"s/regression_\(atx_test_database\)/regress${c03}_\1/g; "\
"s/regress_\(constraint_comments\)/regr${c03}_\1/g; "\
"s/regress_\(no_child_access\)/regr${c03}_\1/g; "\
"s/regress_\(rol_lock1\)/regr${c03}_\1/g; "\
"s/regress_\(nosuper\|view_owner\)/regr${c03}_\1/g; "\
"s/regress_\(truncate_conflict\)/regr${c03}_\1/g; "\
"s/regress_\(alice\|bob\|carol\|dave\)/regr${c03}_\1/g; "\
"s/regress_\([a-z_]*user[0-9]*\)/regr${c03}_\1/g; "\
"s/regress_\([a-z_]*role\)/regr${c03}_\1/g; "\
"s/regress_\([a-z_]*group\)/regr${c03}_\1/g; "\
"s/regress_\(test_[a-z0-9_]*\)/regr${c03}_\1/g; "\
"s/regress_\(priv_[a-z0-9]\+\)/regr${c03}_\1/g; "\
"s/regress_\(rls_[a-z0-9]\+\)/regr${c03}_\1/g; "\
"s/regress_\(passwd[a-z0-9_]*\)/regr${c03}_\1/g; " -i {} \;

done

export ASAN_OPTIONS=detect_leaks=0:abort_on_error=1:disable_coredump=0:strict_string_checks=1:check_initialization_order=1:strict_init_order=1:detect_odr_violation=0

# Prepare master node
$PGBIN/initdb -k -D "$PGDB" --username=postgres
echo "
shared_buffers = 1024MB
max_connections = 1000

log_min_messages = notice
log_min_error_statement = log

#log_connections = on
#log_disconnections = on
log_line_prefix = '%m|%u|%d|%c|'
fsync = off
" >> "$PGDB/postgresql.auto.conf"

# Start server
$PGBIN/pg_ctl -w -t 15 -D "$PGDB" -l server$SID.log start

$PGBIN/createdb test

# Run `make installcheck` in a loop
result=0
time for i in `seq $NUM_ITERATIONS`; do
  echo "iteration $i: `date`"
  for c in `seq $NUM_INSTALLCHECKS`; do
    (cd $SRCCOPYDIR/ic$SID-$c; EXTRA_REGRESS_OPTS="--dbname=regress$c" time make installcheck -C src/test/isolation NO_GENERATED_HEADERS=1 || echo FAIL) >> installcheck$SID-$c.log 2>&1 &
#    (cd $SRCCOPYDIR/ic$SID-$c; EXTRA_REGRESS_OPTS="--dbname=regress$c" time make installcheck-world NO_GENERATED_HEADERS=1 || echo FAIL) >> installcheck$SID-$c.log 2>&1 &
  done
  wait

  echo "installcheck finished: `date`"
  if grep 'test process exited with' installcheck*.log >fail.log || grep 'FAIL' installcheck*.log || grep '_suppression_' server*.log; then
     echo "installcheck FAILED"
     result=1
     break
  fi
  sleep 3
  coredumpctl && result=1
  [ $result -eq 0 ] || break
done

if [ $result -eq 0 ]; then
  # Stop server on success
  $PGBIN/pg_ctl -w -t 90 -D "$PGDB" stop
  echo "ok"
fi
exit $result
